package PScm::Class;

use strict;
use warnings;
use base qw(PScm);
use PScm::Continuation;

sub new {
    my ($class, $parent, $fields, $methods, $env) = @_;

    my $rh_methods = {};
    $class->_populate_methods_hash($rh_methods, $methods);

    return bless {
        parent  => $parent,
        fields  => [$fields->value],
        methods => $rh_methods,
        env     => $env,
    }, $class;
}

sub _populate_methods_hash {
    my ($class, $rh_methods, $methods) = @_;
    if ($methods->is_pair) {
        my $method = $methods->first;
        my ($name, $args, $body) = $method->value;
        $rh_methods->{ $name->value } =
                { args => $args, body => $body };
        $class->_populate_methods_hash($rh_methods, $methods->rest);
    }
}

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;

    $self->make_instance(
        cont {
            my ($new_object) = @_;
            $new_object->call_method(
                $new_object,
                "init", $form, $env,
                cont {
                    my ($dummy, $fail) = @_;
                    $cont->Cont($new_object);
                },
                $fail
            );
        },
        $fail
    );
}

sub make_instance {
    my ($self, $cont, $fail) = @_;

    $self->{parent}->make_instance(cont {
        my ($parent_instance, $fail) = @_;

        $self->{env}->ExtendUnevaluated(
            new PScm::Expr::List(
                PScm::Expr::Symbol->new("class"),    # $self
                PScm::Expr::Symbol->new("super"),    # $parent_instance
                @{ $self->{fields} },                # 0...
            ),
            new PScm::Expr::List(
                $self,                               # "class"
                PScm::Env::Super->new(super => $parent_instance)
                ,                                    # "super"
                ((PScm::Expr::Number->new(0)) x @{ $self->{fields} })
                ,                                    # field...
            ),
            $cont,
            $fail
        );
    });
}

sub get_method {
    my ($self, $method_name, $object) = @_;

    if (exists $self->{methods}{$method_name}) {
        return PScm::Closure::Method->new(
            $self->{methods}{$method_name}{args},
            $self->{methods}{$method_name}{body}, $object);
    }
}

##########################
package PScm::Class::Root;

use base qw(PScm::Class);

sub new {
    my ($class, $env) = @_;

    return bless {
        parent  => 0,
        fields  => [],
        methods => {},
        env     => $env,
    }, $class;
}

sub make_instance {
    my ($self, $cont, $fail) = @_;

    $self->{env}->ExtendUnevaluated(
        new PScm::Expr::Symbol("class"),
        $self,
        $cont,
        $fail
    );
}

1;
