package PScm::Closure;

use strict;
use warnings;
use base qw(PScm);
use PScm::Continuation;

sub new {
    my ($class, $args, $body, $env) = @_;

    bless {
        args => $args,
        body => $body,
        env  => $env,
    }, $class;
}

sub _apply {
    my ($self, $args, $cont, $fail) = @_;

    $self->{env}->ExtendUnevaluated(
        $self->{args},
        $args,
        cont {
            my ($extended_env, $fail) = @_;
            $self->{body}->Eval($extended_env, $cont, $fail);
        },
        $fail
    );
}

sub as_string {
    my ($self) = @_;
    return PScm::Expr::List->new(
        $self->_symbol,
        $self->{args},
        $self->{body}
    )->as_string;
}

################################
package PScm::Closure::Function;

use base qw(PScm::Closure);
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;

    $form->map_eval(
        $env,
        cont {
            my ($evaluated_args, $fail) = @_;
            $self->_apply($evaluated_args, $cont, $fail);
        },
        $fail
    );
}

sub _symbol {
    PScm::Expr::Symbol->new('lambda');
}

#############################
package PScm::Closure::Macro;

use base qw(PScm::Closure);
use PScm::Continuation;

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;

    $self->_apply(
        $form,
        cont {
            my ($new_form, $fail) = @_;
            $new_form->Eval($env, $cont, $fail);
        },
        $fail
    );
}

sub _symbol {
    PScm::Expr::Symbol->new('macro');
}

##############################
package PScm::Closure::Method;

use base qw(PScm::Closure);
use PScm::Continuation;

sub new {
    my ($class, $args, $body, $env) = @_;

    bless {
        args => PScm::Expr::List->Cons(
            PScm::Expr::Symbol->new("this"), $args
        ),
        body => $body,
        env  => $env,
    }, $class;
}

sub ApplyMethod {
    my ($self, $this, $form, $env, $cont, $fail) = @_;

    $form->map_eval(
        $env,
        cont {
            my ($evaluated_args, $fail) = @_;
            $self->_apply(
                PScm::Expr::List->Cons($this, $evaluated_args),
                $cont, $fail);
        },
        $fail
    );
}

1;
