package PScm::Continuation;

use strict;
use warnings;
use base qw(PScm::Expr);

require Exporter;

push our @ISA, qw(Exporter);

our @EXPORT = qw(bounce cont fail);

sub new {
    my ($class, $cont) = @_;
    bless { cont => $cont }, $class;
}

sub cont(&) {
    my ($cont) = @_;
    return PScm::Continuation::Cont->new($cont);
}

sub fail(&) {
    my ($fail) = @_;
    return PScm::Continuation::Fail->new($fail);
}

sub bounce(&) {
    my ($bounce) = @_;
    return PScm::Continuation::Bounce->new($bounce);
}

#################################
package PScm::Continuation::Cont;

use strict;
use warnings;
use base qw(PScm::Continuation);

BEGIN {
    *cont = \&PScm::Continuation::cont;
    *bounce = \&PScm::Continuation::bounce;
}

sub Apply {
    my ($self, $form, $env, $cont, $fail) = @_;
    $form->map_eval(
        $env,
        cont {
            my ($evaluated_args, $fail) = @_;
            $self->Cont($evaluated_args->first, $fail);
        },
        $fail
    );
}

sub Cont {
    my ($self, $arg, $fail) = @_;
    
    bounce { $self->{cont}->($arg, $fail) }
}

#################################
package PScm::Continuation::Fail;

use strict;
use warnings;
use base qw(PScm::Continuation);

BEGIN { *bounce = \&PScm::Continuation::bounce; }

sub Fail {
    my ($self) = @_;
    bounce { $self->{cont}->() }
}

###################################
package PScm::Continuation::Bounce;

use strict;
use warnings;
use base qw(PScm::Continuation);

sub Bounce {
    my ($self) = @_;
    $self->{cont}->();
}

1;
